import re
from datetime import datetime
from typing import Optional

import numpy as np
import pandas as pd

from pyAFL import config
from pyAFL.base.exceptions import LookupError
from pyAFL.session import session


class Season(object):
    """
    A class to represent an AFL season.

    Attributes
    ----------
    year : int
    stats : object
        MatchStats object

    Methods
    -------
    get_season_stats()
        Returns season stats objects, matches, matches_summary, ladders for each round, and a final ladder
    """

    def __init__(self, year: int, url: Optional[str] = None):
        """
        Constructs all the necessary attributes for the Season object.
         - If no season can be found for the given year, a
         `LookupError` exception is raised.

        Parameters
        ----------
            year : int (required)
                which year to get stats for
            url : str (optional)
                url to get stats from
        """

        self.season = year

        if url:
            self.url = url
        else:
            self.url = self._get_season_url()

    def __repr__(self):
        return f"<season: {self.season}>"

    def __str__(self):
        return str(self.season)

    def _get_season_url(self):
        return config.AFLTABLES_STATS_BASE_URL + f"seas/{self.season}.html"

    def get_season_stats(self, force_live=False):
        """
        Returns season stats as per the season stats page
        defined in `self._get_season_url()`

        Parameters
        ----------
            force_live : bool
                If True, does not use cached request

        Returns
        ----------
            stats : obj
                season stats Python object

        """

        resp = session.get(self.url, force_live)
        self._stat_html = resp.text

        try:
            all_dfs = pd.read_html(self._stat_html)
        except ValueError as exc:
            raise LookupError(
                f"Found no season for year {self.season}. Browse https://afltables.com/afl/seas/season_idx.html for a list of all years with data. Year must be an integer."
            ) from exc

        matches = []
        ladders = []
        final_ladder = []
        finals_stage = "Regular season"
        for df in all_dfs:
            # work out what data is in the table represented by this dataframe
            if df.iat[0, 0].startswith("Round"):
                # start of a round
                round = int(df.iat[0, 0].split(" ")[1])
                match = 1  # set to 1 as we're starting a new round
            elif df.shape[0] == 1:
                # Could be denoting which stage of finals, or could be a bye (which we ignore)
                if df.iat[0, 0].endswith("Final"):
                    finals_stage = df.iat[0, 0]
                    match = 1
                    round += 1
            else:
                # it's a match or a ladder
                if df.shape[0] > 1:
                    if df.iat[0, 0].startswith("Rd"):
                        # ladder tables start with Rd N
                        ladders.append(df)
                    elif isinstance(df.columns, pd.MultiIndex):
                        # it's the end-of-season ladder (drop last row of aggregate results which mess up the formatting)
                        final_ladder = df.head(-1)
                    else:
                        # it's a match
                        if not isinstance(df.iat[0, 1], str) and np.isnan(df.iat[0, 1]):
                            # if there are missing values it's because this season hasn't finished yet
                            final_ladder = []
                            break
                        # get the match object for this match
                        current_match = Match(df, round, match, finals_stage)
                        matches.append(current_match)
                        match += 1

        # build a summary dataframe of match data using the following column names and source variable names
        dataframe_fields = {
            "Date": "date",
            "Round": "round",
            "Game number": "game_number",
            "Venue": "venue",
            "Home team": "home_team",
            "Away Team": "away_team",
            "Home team score": "home_team_score",
            "Away team score": "away_team_score",
            "Home team score detail": "home_team_score_detail",
            "Away team score detail": "away_team_score_detail",
            "Winning team": "winning_team",
            "Margin": "margin",
            "Year stage": "finals_stage",
        }
        match_summary = {}
        for df_column_name, match_attribute_name in dataframe_fields.items():
            match_summary[df_column_name] = [
                getattr(match, match_attribute_name) for match in matches
            ]
        match_summary = pd.DataFrame(match_summary)

        ret = SeasonStats(
            season_matches=matches,
            season_ladders=ladders,
            final_ladder=final_ladder,
            match_summary=match_summary,
        )

        return ret


class SeasonStats(object):
    """
    A class to represent AFL season stats.

    Attributes
    ----------
    year : int
        season year
    stats : object
        SeasonStats object

    Methods
    -------
    ...
    """

    def __init__(self, **kwargs):
        """
        Constructs all the necessary attributes for the SeasonStats object.
         - kwargs passed are accessed as class attributes

        """

        super().__init__()
        self.__dict__.update(kwargs)
        pass


class Match:
    """
    A class to represent a match, extracting data from dataframe with match information from AFLtables

    Attributes
    ----------
    date : datetime
    round : int
    game_number : int
    venue : str
    result : str
    winning_team : str
    margin : int
    home_team : str
    away_team : str
    home_team_score : int
    away_team_score : int
    home_team_score_detail : List[int]
    away_team_score_detail : List[int]
    finals_stage : str

    Methods
    -------
    ...
    """

    def __init__(
        self,
        match: pd.DataFrame,
        round: int,
        game_number: int,
        finals_stage: str,
    ):
        """
        Constructs all the necessary attributes for the Match object.

        Parameters
        ----------
            match : DataFrame
                Match record from AFLTables
            round : int
                Round number in the season
            game_number : int
                Game number in the round
            finals_stage : str
                One of 'Regular season', 'Qualifying Final', 'Preliminary Final', 'Elimination Final', 'Grand Final'
        """
        # match is a dataframe that we have to extract data from, example:
        #             0                            1    2                                                                     3
        # 0  West Coast  0.0 Â Â 2.0 Â Â 4.3 Â Â 9.4   58  Fri 15-Apr-2022 5:40 PM (7:40 PM)  Att:  42,888 Venue: Perth Stadium
        # 1      Sydney        5.4 10.10 11.12 18.13  121          Sydney  won by  63 pts  [  Match stats  ]
        #
        # There are variations in the format...

        self.round = round
        self.game_number = game_number
        self.finals_stage = finals_stage
        self.date = self._parse_time(match)
        self.venue = match.iat[0, 3].split("Venue: ")[1].strip()
        self.result = match.iat[1, 3][:-18].strip()
        self.home_team = match.iat[0, 0]
        self.away_team = match.iat[1, 0]
        self.home_team_score = int(match.iat[0, 2])
        self.away_team_score = int(match.iat[1, 2])
        self.home_team_score_detail = self._score_detail(match.iat[0, 1])
        self.away_team_score_detail = self._score_detail(match.iat[1, 1])
        if self.home_team_score > self.away_team_score:
            self.winning_team = self.home_team
            self.margin = self.home_team_score - self.away_team_score
        elif self.home_team_score < self.away_team_score:
            self.winning_team = self.away_team
            self.margin = self.away_team_score - self.home_team_score
        else:
            self.winning_team = ""
            self.margin = 0

    def _parse_time(self, match):
        date_field = match.iat[0, 3].split(" ")
        # remove attendance field, it gets in the way
        if date_field[5] == "Att:":
            date_field = date_field[:4] + date_field[8:]

        # If local time is not AEST, AEST time appears in brackets, we pick AEST time
        if date_field[5].strip() == "Venue:":
            # local time is AEST
            match_date = " ".join(date_field[1:4])
            return datetime.strptime(match_date, "%d-%b-%Y %I:%M %p")
        else:
            match_date = " ".join(date_field[1:2] + date_field[4:6])
            return datetime.strptime(match_date, "%d-%b-%Y (%I:%M %p)")

    def _score_detail(self, score):
        # Score is a string like '0.0 Â Â 2.0 Â Â 4.3 Â Â 9.4' and there are weird characters
        # Get the relevant ones with a regex, then split on decimals to split goals/behinds
        strings = [x.split(".") for x in re.findall("\\d*\\.?\\d+", score)]
        return [int(y) for x in strings for y in x]

    def __repr__(self):
        output = f"MATCH:\n\t{self.home_team} vs {self.away_team}"
        if self.finals_stage != "Regular season":
            output += f" ({self.finals_stage})"
        output += f"\n\tRound: {self.round} game {self.game_number}"
        output += f"\n\tVenue: {self.venue}"
        output += f"\n\tDate: {self.date}"
        output += f"\n\tHome team score: {self.home_team_score_detail}: {self.home_team_score}"
        output += f"\n\tAway team score: {self.away_team_score_detail}: {self.away_team_score}"
        output += f"\n\tResult: {self.result}"
        output += f"\n\t({self.winning_team} {self.margin})\n"
        return output



import sys
import urllib
from datetime import timedelta

from bs4 import BeautifulSoup
from requests_cache import CachedSession
from requests_cache.models import AnyResponse


class __AFLTablesCachedSession(CachedSession):

    def request(
        self,
        method: str,
        url: str,
        *args,
        **kwargs,
    ) -> AnyResponse:
        response = super().request(method, url, **kwargs)
        self.cache.save_response(self.post_process(url, response))
        return response

    def post_process(self, url: str, response: AnyResponse):
        """ Converts the relative urls to absolute urls.
            - Parses the response html
            - Finds all <a> tags
            - If the `href` value is a relative url, prepend it with the current request url.

            Reasoning:
            - afltable.com uses relative urls (eg `../teams/richmond_idx.html` rather
            than `https://afltables.com/afl/teams/richmond_idx.html`).
        """
        soup = BeautifulSoup(response.content, "html.parser")
        for link in soup.findAll("a"):
            if link.attrs.get("href") and "://" not in link.attrs.get("href"):
                link.attrs["href"] = urllib.parse.urljoin(url, link.attrs.get("href"))
        html_out = soup.prettify("utf-8")
        response._content = html_out

        return response

    def get(self, url, force_live=False, **kwargs) -> AnyResponse:
        # If `force_live` kwarg is provided, disable request cache.
        if force_live:
            with self.cache_disabled():
                return super().get(url, **kwargs)

        # Else apply normal request caching logic
        return super().get(url, **kwargs)


session = __AFLTablesCachedSession(
    "test_db" if "pytest" in sys.modules else "pyAFL_html_cache",
    backend="filesystem",
    use_cache_dir=True,                # Save files in the default user cache dir
    cache_control=False,               # Use Cache-Control response headers for expiration, if available
    expire_after=timedelta(days=365),  # Otherwise expire responses after one day
    allowable_codes=[200],             # Cache 400 responses as a solemn reminder of your failures
    allowable_methods=['GET'],         # Cache whatever HTTP methods you want
    stale_if_error=False,               # In case of request errors, use stale cache data if possible
)



import pandas as pd
from bs4 import BeautifulSoup

from pyAFL.base.exceptions import LookupError
from pyAFL.players.models import Player
from pyAFL.session import session


class Team(object):
    """
    A class to represent an AFL Team.

    Attributes
    ----------
    name : str
        team full name.
    players : list
        list of all time players who played for the team (pyAFL player objects).
    games : Pandas DataFrame
        dataframe containing results of all games played by the team.

    """

    def __init__(self, name: str, url_identifier: str):
        """
        Constructs the Team object.

        Parameters
        ----------
            name : str (required)
                name of the person in format "[first] [last]"
            url_identifier : str (required)
                string parameter used in AFLtables URLs to identify team. Note that the naming convention changes from team to team
                Examples: - for Adelaide: url_identifier = "adelaide" (see https://afltables.com/afl/stats/teams/adelaide.html)
                          - for Greater Western Sydney: url_identifier = "gws" (see https://afltables.com/afl/stats/teams/gws.html)
                          - for Western Bulldogs: url_identifier = "bullldogs" (see https://afltables.com/afl/stats/teams/bullldogs.html)

        """

        self.name = name.title()  # Convert to title case for URL string matching
        # URL for all-time-players stats
        self.all_time_players_url = f"https://afltables.com/afl/stats/teams/{url_identifier}.html"
        # URL to all-time-games stats
        self.all_time_games_url = f"https://afltables.com/afl/teams/{url_identifier}/allgames.html"

    def __repr__(self):
        return f"<Team: {self.name}>"

    def __str__(self):
        return self.name

    @property
    def players(self):
        """
        NB - a network request will be made when this class property is called.
        """

        return self._get_players()

    def _get_players(self):
        """
        Returns a list of pyAFL.Player objects for all players contained in `self.all_time_players_url`

        Returns
        ----------
            players : list
                list of pyAFL.Player objects

        """
        resp = session.get(self.all_time_players_url)
        soup = BeautifulSoup(resp.text, "html.parser")

        player_table = soup.find("table")
        player_table_body = player_table.find("tbody")
        player_anchor_tags = player_table_body.findAll("a")

        players = [
            Player(player.text, url=player.attrs.get("href"))
            for player in player_anchor_tags
        ]

        return players

    def season_stats(self, year: int):
        """
        Returns a Pandas dataframe detailing the season stats for the specified year.
        E.g. for Adelaide the table found at https://afltables.com/afl/stats/2020.html#1

        Parameters
        ----------
            year : int (required)
                year as a four-digit integer (e.g. 2019)

        Returns
        ----------
            season_stats : Pandas dataframe
                dataframe summarising individual player (and team total) stats for the specified year.

        """
        season_player_stats_url = f"https://afltables.com/afl/stats/{year}.html"
        resp = session.get(season_player_stats_url)

        if resp.status_code == 404:
            raise Exception(f"Could not find season stats for year: {year}")

        soup = BeautifulSoup(resp.text, "html.parser")
        team_tables = soup.findAll("table")

        for table in team_tables:
            if table.find("th"):
                if self.name in table.find("th").text:
                    df = pd.read_html(str(table))

        if df is None:
            raise LookupError(
                f"Could not find season stats table for team {self.name} in year {year} at URL https://afltables.com/afl/stats/{year}.html"
            )

        season_stats = df[0]
        season_stats.columns = season_stats.columns.droplevel()

        return season_stats

    @property
    def games(self):
        return self._get_games()

    def _get_games(self):
        """
        Returns a Pandas dataframe listing every match contained in `self.all_time_games_url`

        Returns
        ----------
            games : Pandas dataframe
                dataframe listing all games played by the team. Contains results and match metadata.

        """
        resp = session.get(self.all_time_games_url)
        soup = BeautifulSoup(resp.text, "html.parser")

        seasons = soup.findAll("table")

        dfs = []
        for season_html in seasons:
            df = pd.read_html(str(season_html))[0]
            df.columns = df.columns.droplevel(1)
            df = df.iloc[0:-2, :]
            dfs.append(df)

        games = pd.concat(dfs)
        games.index = pd.to_datetime(games.Date)
        games = games.sort_index()

        games = games.rename(
            columns={"A": "Against", "F": "For", "R": "Result", "M": "Margin"}
        )

        return games



import pandas as pd
import pytest

from pyAFL.players.models import Player
from pyAFL.teams import ALL_TEAMS, CURRENT_TEAMS
from pyAFL.teams.models import Team


class TestTeamModel:
    def test_team_name_none_throws_exception(self):
        with pytest.raises(TypeError):
            Team()

    def test_team_property_players(self):
        test_team = ALL_TEAMS[0]
        players = test_team.players

        assert isinstance(players, (list, tuple))
        assert len(players) > 0
        assert isinstance(players[0], Player)
        assert "https://afltables.com/afl/stats/players" in players[0].url

    def test_team_method_season_stats_invalid_year(self):
        test_team = ALL_TEAMS[0]

        with pytest.raises(Exception) as e1:
            test_team.season_stats(123)
        with pytest.raises(Exception) as e2:
            test_team.season_stats(2050)

        assert "Could not find season stats for year" in str(e1)
        assert "Could not find season stats for year" in str(e2)

    def test_team_method_season_stats_valid_year(self):
        test_team = CURRENT_TEAMS[0]
        season_stats = test_team.season_stats(2019)

        assert isinstance(season_stats, pd.DataFrame)
        assert "Player" in season_stats.columns

    def test_team_property_games(self):
        test_team = ALL_TEAMS[0]
        games = test_team.games

        assert isinstance(games, pd.DataFrame)
        assert isinstance(games.index, pd.DatetimeIndex)
        assert len(games.index) > len(games.loc["2019-01-01":"2019-08-01"].index)



import re

import pandas as pd
from bs4 import BeautifulSoup

from pyAFL import config
from pyAFL.base.exceptions import LookupError
from pyAFL.session import session


class Player(object):
    """
    A class to represent an AFL player.

    Attributes
    ----------
    name : str
        first name of the person
    stats : object
        PlayerStats object

    Methods
    -------
    ...
    """

    def __init__(self, name: str, url: str = None, team: str = None):
        """
        Constructs all the necessary attributes for the Player object.
         - If `name` returns two or more players, the (optional) parameter
         "team" is used to select the correct player.
         - If no player can be found for the given "name", "team"
         combination, a `LookupError` exception is raised.

        Parameters
        ----------
            name : str (required)
                name of the person in format "[first] [last]"
            team : str (string)
                name of team that the player has played in during their career
        """

        self.name = name.title()  # Convert to title case for URL string matching
        self.name = self.name.replace("\n", "").strip()
        if url:
            self.url = url
        else:
            self.url = self._get_player_url()

    def __repr__(self):
        return f"<Player: {self.name}>"

    def __str__(self):
        return self.name

    def _get_player_url(self):
        last_initial = self.name.split(" ")[1][0]
        player_list_url = (
            config.AFLTABLES_STATS_BASE_URL + f"stats/players{last_initial}_idx.html"
        )

        resp = session.get(player_list_url)
        soup = BeautifulSoup(resp.text, "html.parser")

        url_list = soup.findAll(
            "a",
            href=re.compile(
                f"players/{self.name[0]}/{self.name.replace(' ', '_')}", re.I
            ),
        )

        # If no matches found, raise LookupError
        if len(url_list) == 0:
            raise LookupError(
                f"Found no players with name {self.name}. Browse https://afltables.com/afl/stats/playersA_idx.html for a list of all players. Name must be in format '[first] [last]'."
            )

        # If more than one name is matched, print warning message and return first.
        if len(url_list) > 1:
            print(
                f"Warning: {len(url_list)} players have been found for name: {self.name}. Returning only the first"
            )

        return url_list[0].attrs.get("href")

    def get_player_stats(self):
        """
        Returns player stats as per the player stats page defined in `self._get_player_url()`

        Returns
        ----------
            stats : obj
                player stats Python object

        """

        resp = session.get(self.url)
        self._stat_html = resp.text

        soup = BeautifulSoup(self._stat_html, "html.parser")

        all_dfs = pd.read_html(self._stat_html)
        season_dfs = pd.read_html(self._stat_html, match=r"[A-Za-z]* - [0-9]{4}")

        season_stats_total = all_dfs[0]  # The first table on the page
        season_stats_average = all_dfs[1]  # The second table on the page

        ret = PlayerStats(
            season_stats_total=season_stats_total,
            season_stats_average=season_stats_average,
            season_results=season_dfs,
        )

        return ret


class PlayerStats(object):
    """
    A class to represent an AFL player.

    Attributes
    ----------
    season_stats_total : object
    season_stats_average : object
    season_results : object

    Methods
    -------
    ...
    """

    def __init__(self, **kwargs):
        """
        Constructs all the necessary attributes for the PlayerStats object.
         - kwargs passed are accessed as class attributes

        """

        super().__init__()
        self.__dict__.update(kwargs)
        pass



import pytest

from pyAFL.base.exceptions import LookupError
from pyAFL.players.models import Player, PlayerStats


class TestPlayerModel:
    def test_player_name_none_throws_exception(self):
        with pytest.raises(TypeError):
            Player()

    def test_player_classmethod_get_player_url_success(self):
        assert (
            Player("Nathan Brown")._get_player_url() ==
            "https://afltables.com/afl/stats/players/N/Nathan_Brown0.html"
        )
        assert (
            Player("Nick Riewoldt")._get_player_url() ==
            "https://afltables.com/afl/stats/players/N/Nick_Riewoldt.html"
        )
        assert (
            Player("Tony Lockett")._get_player_url() ==
            "https://afltables.com/afl/stats/players/T/Tony_Lockett.html"
        )
        assert (
            Player("Duncan MacGregor")._get_player_url() ==
            "https://afltables.com/afl/stats/players/D/Duncan_MacGregor.html"
        )

    def test_player_classmethod_get_player_url_failure(self):
        with pytest.raises(LookupError) as e:
            Player("Babe Ruth")._get_player_url()

        assert "Found no players with name" in str(e)

    def test_player_classmethod_get_player_stats(self):
        player = Player("Nick Riewoldt")

        assert isinstance(player.get_player_stats(), PlayerStats)



from pyAFL.teams.models import Team

ADE = Team("Adelaide", "adelaide")
BRI = Team("Brisbane Lions", "brisbanel")
BRB = Team("Brisbane Bears", "brisbaneb")
CAR = Team("Carlton", "carlton")
COL = Team("Collingwood", "collingwood")
ESS = Team("Essendon", "essendon")
FIZ = Team("Fitzroy", "fitzroy")
FRE = Team("Fremantle", "fremantle")
GEE = Team("Geelong", "geelong")
GC = Team("Gold Coast", "goldcoast")
GWS = Team("Greater Western Sydney", "gws")
HAW = Team("Hawthorn", "hawthorn")
MEL = Team("Melbourne", "melbourne")
NOR = Team("North Melbourne", "kangaroos")
POR = Team("Port Adelaide", "padelaide")
RIC = Team("Richmond", "richmond")
STK = Team("St Kilda", "stkilda")
SYD = Team("Sydney", "swans")
UNI = Team("University", "university")
WCE = Team("West Coast", "westcoast")
WBD = Team("Western Bulldogs", "bullldogs")

ALL_TEAMS = [
    ADE,
    BRI,
    BRB,
    CAR,
    COL,
    ESS,
    FIZ,
    FRE,
    GEE,
    GC,
    GWS,
    HAW,
    MEL,
    NOR,
    POR,
    RIC,
    STK,
    SYD,
    UNI,
    WCE,
    WBD,
]

CURRENT_TEAMS = [
    ADE,
    BRI,
    CAR,
    COL,
    ESS,
    FRE,
    GEE,
    GC,
    GWS,
    HAW,
    MEL,
    NOR,
    POR,
    RIC,
    STK,
    SYD,
    WCE,
    WBD,
]


import pytest
import requests
from requests_cache.models import CachedResponse, Response

from pyAFL.session import session


class TestRequestCaching:

    @pytest.fixture(autouse=True, scope="function")
    def before_each(self):
        session.cache.clear()

    def test_none_url_throws_exception(self):
        with pytest.raises(TypeError):
            session.get()

    def test_missing_url_schema_throws_exception(self):
        with pytest.raises(requests.exceptions.MissingSchema):
            session.get("abcdefgh")

    def test_invalid_url_throws_exception(self):
        with pytest.raises(requests.exceptions.ConnectionError):
            session.get("https://abcdefgh")

    def test_second_request_is_from_cache(self):
        url = "https://afltables.com/afl/stats/playersA_idx.html"

        resp1 = session.get(url)
        assert isinstance(resp1, Response)
        assert hasattr(resp1, "from_cache")
        assert not resp1.from_cache

        resp2 = session.get(url)
        assert isinstance(resp2, CachedResponse)
        assert hasattr(resp2, "from_cache")
        assert resp2.from_cache

    def test_all_requests_with_force_live_are_not_from_cache(self):
        url = "https://afltables.com/afl/stats/playersB_idx.html"

        resp1 = session.get(url, force_live=True)
        assert isinstance(resp1, Response)
        assert hasattr(resp1, "from_cache")
        assert not resp1.from_cache

        resp2 = session.get(url, force_live=True)
        assert isinstance(resp2, Response)
        assert hasattr(resp2, "from_cache")
        assert not resp2.from_cache

        resp3 = session.get(url, force_live=True)
        assert isinstance(resp3, Response)
        assert hasattr(resp3, "from_cache")
        assert not resp3.from_cache



class AFLObject(object):
    def _get_object_from_db(self, **kwargs):
        """
        Searches local DB for queried object.

        Parameters
        ----------
            **kwargs : query terms

        Returns
        -------
            Database object if found;
            else None
        """

        return None


import setuptools

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="pyAFL",
    version="0.4.1",
    description="Python data fetching library for the Australian Football League",
    long_description="pyAFL is a AFL (Australian Football League) data fetching libary. It scrapes data from https://afltables.com/ and converts results to structured Python objects for easier analytics.",
    url="https://github.com/RamParameswaran/pyAFL",
    author="Ram Parameswaran",
    author_email="ram@findram.dev",
    license="MIT",
    packages=setuptools.find_packages(
        exclude=[
            "tests",
        ]
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=required,
)



from datetime import datetime

import pytest

from pyAFL.base.exceptions import LookupError
from pyAFL.seasons.models import Season, SeasonStats


class TestSeasonModel:
    def test_season_none_throws_exception(self):
        with pytest.raises(TypeError):
            Season()

    def test_season_classmethod_get_season_url_success(self):
        assert (
            Season(2019)._get_season_url() == "https://afltables.com/afl/seas/2019.html"
        )

    def test_season_classmethod_get_season_url_failure(self):
        with pytest.raises(LookupError) as e:
            Season(12).get_season_stats()

        assert "Found no season for year" in str(e)

    def test_season_classmethod_get_season_stats(self):
        season = Season(2017)

        assert isinstance(season.get_season_stats(), SeasonStats)


class TestSeasonStats:
    season = Season(2017)
    stats = season.get_season_stats()
    matches = stats.season_matches
    # Match 54
    # St Kilda	                2.4   6.7  10.9 16.12	108
    # Greater Western Sydney	4.2   7.6 11.10 12.13	85
    # Fri 05-May-2017 7:50 PM Att: 21,160 Venue: Docklands
    # St Kilda won by 23 pts

    # Match 57
    # Port Adelaide	  2.3   5.7  7.13 12.15	87
    # West Coast	  4.1   7.3  12.4  15.7	97
    # Sat 06-May-2017 4:05 PM (4:35 PM) Att: 38,333 Venue: Adelaide Oval
    # West Coast won by 10 pts [Match stats]

    def test_season_match_parsing_home_team(self):
        assert self.matches[54].home_team == "St Kilda"

    def test_season_match_parsing_away_team(self):
        assert self.matches[54].away_team == "Greater Western Sydney"

    def test_season_match_parsing_time_AEST(self):
        assert self.matches[54].date == datetime.strptime(
            "05-May-2017 (7:50 pm)", "%d-%b-%Y (%I:%M %p)"
        )

    def test_season_match_parsing_time_not_AEST(self):
        assert self.matches[57].date == datetime.strptime(
            "06-May-2017 (4:35 pm)", "%d-%b-%Y (%I:%M %p)"
        )

    def test_season_match_home_team_score(self):
        assert self.matches[54].home_team_score == 108

    def test_season_match_result(self):
        # Important whitespace
        assert self.matches[54].result == "St Kilda  won by  23 pts"

    def test_season_match_margin(self):
        assert self.matches[54].margin == 23

    def test_match_repr(self):
        output = f"MATCH:\n\tPort Adelaide vs West Coast"
        output += f"\n\tRound: 7 game 4"
        output += f"\n\tVenue: Adelaide Oval"
        output += f"\n\tDate: 2017-05-06 16:35:00"
        output += f"\n\tHome team score: [2, 3, 5, 7, 7, 13, 12, 15]: 87"
        output += f"\n\tAway team score: [4, 1, 7, 3, 12, 4, 15, 7]: 97"
        output += f"\n\tResult: West Coast  won by  10 pts"
        output += f"\n\t(West Coast 10)\n"
        assert repr(self.matches[57]) == output

    def test_finals_match_repr(self):
        output = f"MATCH:\n\tGeelong vs Richmond (Qualifying Final)"
        output += f"\n\tRound: 25 game 1"
        output += f"\n\tVenue: M.C.G."
        output += f"\n\tDate: 2017-09-08 19:50:00"
        output += f"\n\tHome team score: [0, 4, 2, 4, 4, 9, 5, 10]: 40"
        output += f"\n\tAway team score: [2, 4, 3, 7, 6, 10, 13, 13]: 91"
        output += f"\n\tResult: Richmond  won by 51 pts"
        output += f"\n\t(Richmond 51)\n"
        assert repr(self.matches[199]) == output
